// volumetric renderer demo
// author: info@toxi.co.uk
// last modified: 12/09/03

// progressively renders volumetric data (voxels) using a simple ray marching implementation
// more data sets can be downloaded from: http://www.volvis.org/

// click and drag to rotate view
// keyboard controls:
// - and + to zoom in/out
// 1-9 to change contrast

// user defined constants
int MAX_RESOLUTION = 16;
float RAY_PRECISION = 5;

Volume vol;
int res=MAX_RESOLUTION;
boolean clicked=false;
int clickX,clickY;
int currZoom=500;
int currContrast=48;

void setup() {
  size(400,400);
  noBackground();
  noStroke();
  vol=new Volume("foot.raw.gz",256,256,256);
  vol.loadColours("gradient.jpg");
  vol.setView(275,0,0,currZoom);
}

void loop() {
  vol.focus=currZoom;
  if (clicked) {
    vol.rotx=vol.rotx+(int)((clickY-mouseY)*0.025);
    vol.roty=vol.roty+(int)((clickX-mouseX)*0.025);
    if (vol.rotx<0) vol.rotx+=360;
    if (vol.roty<0) vol.roty+=360;
    println(vol.rotx+" "+vol.roty);
    res=MAX_RESOLUTION;
    vol.render(res,currContrast);
  } else if (res>1) {
    res=max(res/2,1);
    vol.render(res,currContrast);
  }
}

void mousePressed() {
  clicked=true;
  clickX=mouseX;
  clickY=mouseY;
}

void mouseReleased() {
  clicked=false;
}

void keyPressed() {
  if (key=='-' || key=='_') { currZoom=max(currZoom-10,200); res=MAX_RESOLUTION*2; }
  if (key=='+' || key=='=') { currZoom=min(currZoom+10,1200); res=MAX_RESOLUTION*2; }
  if (key>='1' && key<='9') { currContrast=160-(key-48)*16; res=MAX_RESOLUTION*2; }
}

// volumetric renderer class

class Volume {
  int w,h,d;
  int w2,h2,d2;
  int sz;
  byte[] data;
  float radius,r2;
  Vec3d origin;
  int rotx,roty,rotz;
  int focus;
  float isectC;
  float begin,end;
  BImage colourMap;
  
  // constructor
  // specify filename of voxel file and the space's width, height, depth
  // NOTE: this code will only work with 8bit voxels
  Volume(String fname, int ww, int hh, int dd) {
    w=ww;
    h=hh;
    d=dd;
    // offsets at the center/origin of the voxel space
    w2=w/2;
    h2=h/2;
    d2=d/2;
    // compute radius of the bounding sphere of the volume
    radius=(int)(new Vec3d(w2,h2,d2)).magnitude();
    // precalc constants for intersection check with bounding sphere
    origin=new Vec3d(0,-400,0);
    isectC=4*((origin.x*origin.x + origin.y*origin.y + origin.z*origin.z)-radius*radius);
    // load voxel data
    data=loadBytes(fname);
    println(fname+" loaded: "+data.length+" bytes");
  }

  // loads an image of (256 x 1) as colour map to colourcode voxel densities
  void loadColours(String fname) {
    colourMap=loadImage(fname);
  }
  
  // set new view angle/zoom
  void setView(int rx,int ry, int rz, int foc) {
    rotx=rx;
    roty=ry;
    rotz=rz;
    focus=foc;
  }

  // get voxel value at specified position
  // return 0 if outside bounds
  int getVoxelAt(int x,int y, int z) {
    return (x>=0 && x<w && y>=0 && y<h && z>=0 && z<d) ? (int)data[(z<<16)+(y<<8)+x] : 0;
  }

  // render volume with the current view settings
  // use resolution parameter to control level of detail
  // the renderer work by sending rays through the voxel space and adding up
  // the density values of voxels hit on the way. the total sum is then used as pixel value,
  // colour coded via the loaded colourMap image
  // for more detail see comments for marchRay() method
  void render(int res,int contrast) {
    int col=0;
    Vec3d rayDir=new Vec3d();
    int scrW2=width/2;
    int scrH2=height/2;
    for(int i=0; i<width; i+=res) {
      for(int j=0; j<height; j+=res) {
        rayDir.x=-(j-scrH2);
        rayDir.y=focus;
        rayDir.z=i-scrW2;
        rayDir.normalize();
        col=colourMap.pixels[min(marchRay(rayDir,RAY_PRECISION,contrast),0xff)];
        fill(col>>16&0xff,col>>8&0xff,col&0xff);
        rect(i,j,res,res);
      }
      if (clicked && res<MAX_RESOLUTION) break;
    }
  }

  // simple check if ray actually will penetrate/intersect the volume
  // this is to speed up rendering as this function also computes
  // the start and end scale of the ray, so if the ray intersects the volume only
  // on the edge, less voxels will be checked than if ray goes straight through the
  // centre of the voxel space...
  // code based on paul bourke's C implementation here:
  // http://astronomy.swin.edu.au/~pbourke/geometry/sphereline/
  boolean intersectsBoundingSphere(float dirx,float diry, float dirz) {
    float B=0,D=0;
    B = 2 * (dirx*origin.x + diry*origin.y + dirz*origin.z);
    D = B*B - isectC;
    begin=end=0;
    if(D>0.0) {
      float rootD=sqrt(D);
      begin = (-B + rootD) * 0.5;
      end = (-B - rootD) * 0.5;
      if(begin>0.0) {
        float tmp = end;
        end = begin;
        begin = tmp;
      }
    }
    return (begin<0.0) ? true : false;
  }

  // main ray marching method
  // examines voxel space for voxels hit by specified ray
  // the returned value will be the sum of the density values of all by the ray's impacted voxels
  // uses trilinear filtering instead of linear interpolation for smoother results
  // interpolation is needed as ray resolution can be (and usually is) higher than voxel resolution
  // also makes use of 16bit fixed point math for speed reasons
  int marchRay(Vec3d dir,float inc, int contr)
  {
    int raySum=0;
    int xx,yy,zz,xx1,yy1,zz1;
    Vec3d startPos=new Vec3d(origin);
    float currX,currY,currZ;
    int fracX,fracY,fracZ,ifracX;
    int vUL,vUR,vLL,vLR;
    int s1,s2;
    int sliceSum1,sliceSum2;
    int contrast = (int)(contr/inc*0x10000);
    
    if( !intersectsBoundingSphere(-dir.x,-dir.y,-dir.z) ) return 0;

    // transform ray's start position and its direction
    applyView(startPos,roty,rotz,rotx);
    applyView(dir,roty,rotz,rotx);
    
    for(float i=-begin; i<-end; i+=inc) {
      currX = startPos.x + dir.x*i;
      currY = startPos.y + dir.y*i;
      currZ = startPos.z + dir.z*i;

      xx=(int)currX;
      yy=(int)currY;
      zz=(int)currZ;
      // get the fractional parts of the voxel coordinates
      // e.g. the frac part of 23.42 = 0.42...
      // by multiplying this with 0x1000 (or 65536 decimal) we effectively
      // move the result 16 bits to the left and can work purely with integers
      // in 16bit fixed point math 65536 = 1, 32768 = 0.5, 16384 = 0.25 etc.
      // the most important step is to shift the result back before returning to normal number formats
      // this is achieved by dividing the result by 65536 again at the end
      fracX = (int)((currX-xx)*0x10000); ifracX=0x10000-fracX;
      fracY = (int)((currY-yy)*0x10000);
      fracZ = (int)((currZ-zz)*0x10000);

      xx+=w2; yy+=h2; zz+=d2;
      xx1=xx-1; yy1=yy-1; zz1=zz-1;
      
      // trilinear filtering
      // first interpolate the 4 neighbour voxels in the previous Z slice...
      vUL=getVoxelAt(xx1,yy1,zz1);
      vUR=getVoxelAt(xx,yy1,zz1);
      vLR=getVoxelAt(xx,yy,zz1);
      vLL=getVoxelAt(xx1,yy,zz1);
      s1 = (vUL*ifracX+vUR*fracX)>>16;
      s2 = (vLL*ifracX+vLR*fracX)>>16;
      sliceSum1 = (s1*(0x10000-fracY)+s2*fracY)>>16;
      // ...then interpolate the 4 neighbour voxels in the current Z slice
      vUL=getVoxelAt(xx1,yy1,zz);
      vUR=getVoxelAt(xx,yy1,zz);
      vLR=getVoxelAt(xx,yy,zz);
      vLL=getVoxelAt(xx1,yy,zz);
      s1 = (vUL*ifracX+vUR*fracX)>>16;
      s2 = (vLL*ifracX+vLR*fracX)>>16;
      sliceSum2 = (s1*(0x10000-fracY)+s2*fracY)>>16;

      // finally add the interpolated value between both slice sums to the total ray sum
      raySum+= abs(sliceSum1*(0x10000-fracZ)+sliceSum2*fracZ);
    }
    return raySum/contrast;
  }

  // apply view port to the specified vector
  void applyView(Vec3d p,int angleX, int angleY, int angleZ) {
    // Z rotation, only affects XY plane
    angleZ%=360;
    float s = g.sinLUT[angleZ];
    float c = g.cosLUT[angleZ];
    float t = p.x*c - p.y*s;
    p.y = p.y*c + p.x*s;
    p.x = t;
    // X rotation (YZ plane only)
    angleX%=360;
    s = g.sinLUT[angleX];
    c = g.cosLUT[angleX];
    t = p.y*c - p.z*s;
    p.z = p.z*c + p.y*s;
    p.y = t;
    // Y rotation (XZ plane only)
    angleY%=360;
    s = g.sinLUT[angleY];
    c = g.cosLUT[angleY];
    t = p.x*c + p.z*s;
    p.z = p.z*c - p.x*s;
    p.x = t;
  }
}

// very simple 3d vector class
class Vec3d {
  float x,y,z;
  Vec3d() { x=y=z=0; }
  Vec3d(Vec3d v) { x=v.x; y=v.y; z=v.z; }
  
  Vec3d(int xx,int yy, int zz) {
    x=(float)xx; y=(float)yy; z=(float)zz;
  }
  Vec3d(float xx,float yy, float zz) {
    x=xx; y=yy; z=zz;
  }
  void normalize() {
    float mag=magnitude();
    if (mag>0.0) {
      x/=mag; y/=mag; z/=mag;
    }
  }
  float magnitude() {
    return sqrt(x*x+y*y+z*z);
  }
}



